{-# OPTIONS -cpp -fglasgow-exts #-}

module Cell (
  Cell(..),
  Threshold(..),
  emptyCell,
  bestParse,
  costCell,
  rootCost,
  accumulateCell,
  chopCell,
  litToCell
) where

import Context
import Util
import Cost
import Parse
import Lex

newtype Cell = Cell [Costed ParseTree] deriving (Eq, Show)

data Threshold = Threshold# { maxDiff :: !Cost, maxLen :: !Int }

emptyCell :: Cell
emptyCell = Cell []

-- Maybe we should optimize & cache this in a cell as well...
-- See what the profiler says.
rootCost :: Costed ParseTree -> Cost
rootCost t = cost t

localCost :: Costed ParseTree -> Cost
localCost t = cost t -$ priorCost (context (value t))

bestParse :: Threshold -> Cell -> Costed ParseTree
bestParse _ (Cell (c:cs)) = c

costCell :: Threshold -> Cell -> Costed Cell
costCell thresh c = Costed# (cost (bestParse thresh c)) c

{-
 You can think of 'accumulateCell thresh x a b' as a kind of +=*:
   ""x += a*b""
 More precisely, merge into x the good ways of combining the parse
 trees in a and b
-}
accumulateCell :: Threshold -> Cell -> Cell -> Cell -> Cell
accumulateCell thresh x a b =
  foldl (mergeCell thresh)  x (genMerges (chopCell thresh a)
			                 (chopCell thresh b))

-- Alias: merge, with a cutoff
mergeCell :: Threshold -> Cell -> Cell -> Cell
mergeCell _ (Cell []) c2 = c2
mergeCell _ c1 (Cell []) = c1
mergeCell (Threshold# maxDiff _) (Cell (x:xs)) (Cell (y:ys)) =
   let cx = rootCost x
       cy = rootCost y
       merge' lim cx (x:xs) ys | cx >= lim = ys
       merge' _   _  xs     []             = xs
       merge' lim cx (x:xs) (y:ys) =
           let cy = rootCost y
           in if (cx < cy) then x:(merge' lim cy (y:ys) xs)
                           else y:(merge' lim cx (x:xs) ys)
   in if (cx < cy) then Cell (x:merge' (cx+$maxDiff) cy (y:ys) xs)
                   else Cell (y:merge' (cy+$maxDiff) cx (x:xs) ys)

-- Currently, the cells are chopped independently before entering genMerges.
-- Maybe better would be to look a balanced difference down the combined list.
-- But that's harder, so I'll consider it later.
genMerges :: Cell -> Cell -> [Cell]
genMerges (Cell xs) (Cell ys) =
  map Cell [join x y | x <- xs, y <- ys]
  where join :: Costed ParseTree -> Costed ParseTree -> [Costed ParseTree]
        join x y = [Costed# (localCost x +$ localCost y +$ cost +$ priorCost cntxt)
		            (Branch cntxt [value x, value y])
		    | PF cntxt cost <- productionsFor (context (value x))
		                                      (context (value y))]

chopCell :: Threshold -> Cell -> Cell
chopCell _ (Cell []) = emptyCell
chopCell (Threshold# maxDiff maxLen) (Cell (c:cs)) =
   Cell (c:let best = rootCost c in
           takeWhile (\c -> rootCost c < maxDiff)
                     (take maxLen cs))

litToCell :: (ContextCombo, Literal) -> Cell
litToCell v@(cntxt, (_, False)) = Cell [Costed# (priorCost cntxt) (PLit v)]
litToCell (cntxt, l@(_, True)) = Cell (map (\c -> Costed# (priorCost c) (PLit (c,l)))
				           (spaceCombos cntxt))
