module Parse (
  ParseTree(..),
  flatten,
  parse,
  unParse,
  context
) where
import Prelude hiding (lex)
import Lex
import Tag
import Maybe
import Context
import Util

type LitContext = (ContextCombo, Literal)
data ParseTree = Branch ContextCombo [ParseTree] |
	         PLit LitContext
  deriving (Eq, Show)

context :: ParseTree -> ContextCombo
context (Branch cntxt _) = cntxt
context (PLit (cntxt, _)) = cntxt

naiveParse :: [Lexeme] -> ParseTree
naiveParse lexs = fst (naiveParse' (unitCombo rootContext) [] lexs)

naiveParse' cntxt l (Open t:r) =
  let (tree, Close _:r') = naiveParse' (applyTagContextCombo t cntxt) [] r in
  naiveParse' cntxt (tree:l) r'
naiveParse' cntxt l (Lit lit:r) = naiveParse' cntxt (PLit (cntxt,lit):l) r
naiveParse' cntxt l r = (Branch cntxt (reverse l), r)

flatten :: ParseTree -> [LitContext]
flatten t = flatten' t []

flatten' (PLit x) r = x:r
flatten' (Branch _ ts) r = foldr flatten' r ts 

{-
Normalize a parse tree:
o Make all branches have valence at least 2
o Merge adjacent literals with the same context
o Merge spaces left or right, as possible
-}
normalizeTree :: ParseTree -> ParseTree
normalizeTree t = fromJust (fst (normalizeTree' t (mergeLiterals (flatten t))))

normalizeTree' :: ParseTree -> [Maybe LitContext] -> (Maybe ParseTree, [Maybe LitContext])
normalizeTree' (PLit _) (Nothing:lcs) = (Nothing, lcs)
normalizeTree' (PLit _) (Just lc:lcs) = (Just (PLit lc), lcs)
normalizeTree' (Branch cntxt ts) lcs =
  let (ts', lcs') = thread normalizeTree' ts lcs
      ts'' = catMaybes ts'
  in case ts'' of
       []  -> (Nothing, lcs')
       [t] -> (Just t, lcs')
       _   -> (Just (Branch cntxt ts''), lcs')

{-
mergeLiterals:
Combine literals that can be combined
-}
mergeLiterals :: [LitContext] -> [Maybe LitContext]
mergeLiterals [] = []
mergeLiterals ((c1, (l1, b1)):lcs) = tryMergeR (c1, b1) [l1] lcs

tryMergeR (c1, b1) ls ((c2, (l2, b2)):lcs) | mergeable (c1, b1) (c2, b2) =
  Nothing:tryMergeR (c2, b2) (l2:ls) lcs
tryMergeR (c1, b1) ls lcs =
  tryMergeL (c1,b1) ls 0 lcs

tryMergeL (c1, b1) ls n ((c2, (l2, b2)):lcs) | mergeable (c2, b2) (c1, b1) =
  tryMergeL (c1, b1) (l2:ls) (n+1) lcs
tryMergeL (c1, b1) ls n lcs =
  let ls' = concat (reverse ls)
      ls'' = if (isTT c1) then ls' else mergeSpaces ls'
  in Just (c1, (ls'', b1)):replicate n Nothing ++ mergeLiterals lcs

mergeable :: (ContextCombo, Bool) -> (ContextCombo, Bool) -> Bool
mergeable (cntxt1, False) (cntxt2, False)
  | cntxt1 == cntxt2 = True
mergeable (cntxt1, True) (cntxt2, _)
  | sameSpaceContextCombo cntxt1 cntxt2 = True
mergeable _ _ = False

makeBinary :: ParseTree -> ParseTree
makeBinary t@(PLit _) = t
makeBinary t@(Branch c [t1,t2]) = Branch c [makeBinary t1,
					    makeBinary t2]
makeBinary (Branch c (t1:ts)) = Branch c [makeBinary t1,
					  makeBinary (Branch c ts)]

parse :: [Lexeme] -> ParseTree
parse = makeBinary . normalizeTree . naiveParse

unParse :: ParseTree -> [Lexeme]
unParse t = unParses rootContext t []

unParses c1 (PLit (c2, lit)) l =
  let (ts, c2') = pathTo c1 c2 in
  showOpen ts (Lit lit:showClose ts l)
unParses c1 (Branch c2 trees) l =
  let (ts, c2') = pathTo c1 c2 in
  showOpen ts (unParses' c2' trees (showClose ts l))

unParses' :: Context -> [ParseTree] -> [Lexeme] -> [Lexeme]
unParses' c [] = id
unParses' c (t:ts) = unParses c t . unParses' c ts

showOpen [] = id
showOpen (t:ts) = (Open t:) . showOpen ts

showClose ts = showClose' (reverse ts)
showClose' [] = id
showClose' (t:ts) = (Close t:) . showClose' ts
