{-# OPTIONS -fglasgow-exts #-}

module Testing where

import PrelGHC
import Validate
import Tag (Tag(TEM, TPL, TTT), allTags)
import qualified Context
import qualified Cost
import PrelBase (Int(I#))
import List (sort, sortBy)
import Exception (try, throw, evaluate, Exception(..))
import IOExts (unsafePerformIO)
import Cost (zeroCost, (+$))
import Set (setToList)

errorToMaybe :: a -> Maybe a
errorToMaybe x = unsafePerformIO $ do
    y <- try (evaluate x)
    case y of Left (ErrorCall        _) -> return Nothing
              Left (PatternMatchFail _) -> return Nothing
              Left exception            -> throw exception
              Right value               -> return (Just value)

encodeCost :: Cost -> Cost.Cost
encodeCost (I# i#) = Cost.Cost# i#

comboToList :: ContextCombo -> [Context]
comboToList (ContextCombo m s c)
    = [ Context m' s' c'
      | m' <- comboToListM m
      , s' <- comboToListUpd s
      , c' <- comboToListUpd c ]

comboToListM (MinimizeSet e u v)
    = [ Minimize e u v'
      | v' <- comboToListV v ]

comboToListV (Weak    ) = [Base, EM]
comboToListV (Simply v) = [v]

comboToListUpd :: (Update a, Finite a) => RootedSet a -> [Rooted a]
comboToListUpd Nothing = [Root, Spec (head everything)]
    -- just testing, so we only take 2
comboToListUpd (Just Root) = [Root]
comboToListUpd (Just (Spec set)) = map Spec (take 2 (setToList set))
    -- again, just testing (so we only take 2)

-- List values are lists of failurse
-- Boolean values are true if the test failed

testCoding :: [Context]
testCoding =
    [ c
    | c <- everything
    , case encodeContext c of { ec ->
      case decodeContext ec of { dec ->
      case encodeContext dec of { edec ->
      dec /= c || edec /= ec } } } ]

testProductionsFor :: [(ContextCombo, ContextCombo)]
testProductionsFor =
    [ (c1, c2)
    | c1 <- everything
    , let ec1 = encodeContextCombo c1
    , c2 <- everything
    , let ec2 = encodeContextCombo c2
    , let slow = [ Context.PF (encodeContextCombo c0) (encodeCost cCost)
                 | (c0, cCost) <- productionsFor c1 c2 ]
    , let fast = Context.productionsFor ec1 ec2
    , sortBy cmp slow /= sortBy cmp fast ]
    where cmp (Context.PF (Context.ContextCombo# x#) _)
              (Context.PF (Context.ContextCombo# y#) _)
            | x# `ltWord#` y# = LT
            | x# `gtWord#` y# = GT
            | otherwise       = EQ

testProductionsForSorting :: [(Context.ContextCombo, Context.ContextCombo)]
testProductionsForSorting =
    [ (c1, c2)
    | c1 <- everything'
    , c2 <- everything'
    , let list = Context.productionsFor c1 c2
    , or (zipWith wrongOrder list (tail list)) ]
    where everything' = map encodeContextCombo everything
          wrongOrder (Context.PF _ cost1) (Context.PF _ cost2) = cost1 > cost2

testPathTo :: [(Context, ContextCombo)]
testPathTo =
    [ (c0, c1)
    | c0 <- everything
    , let ec0 = encodeContext c0
    , c1 <- everything
    , let ec1 = encodeContextCombo c1
    , let slow = errorToMaybe ( case pathTo c0 c1 of { (t,r) ->
                                length t `seq` r `seq` (t, encodeContext r) } )
    , let fast = errorToMaybe (Context.pathTo ec0 ec1)
    , slow /= fast ]

testCodingCombo :: [ContextCombo]
testCodingCombo =
    [ c
    | c <- everything
    , case encodeContextCombo c of { ec ->
      case decodeContextCombo ec of { dec ->
      case encodeContextCombo dec of { edec ->
      dec /= c || edec /= ec } } } ]

testUnitCombo :: [Context]
testUnitCombo =
    [ c
    | c <- everything
    , encodeContextCombo (unitCombo c) /= Context.unitCombo (encodeContext c) ]

testRoot :: Bool
testRoot = encodeContext rootContext /= Context.rootContext

testPriorCost :: [ContextCombo]
testPriorCost =
    [ c
    | c <- everything
    , encodeCost (priorCost c) /= Context.priorCost (encodeContextCombo c) ]

testApplyTagContext :: [(Tag, Context)]
testApplyTagContext =
    [ (t, c)
    | c <- everything
    , let ec = encodeContext c
    , t <- allTags
    , encodeContext (applyTagContext t c) /= Context.applyTagContext t ec ]

testApplyTagContextCombo :: [(Tag, ContextCombo)]
testApplyTagContextCombo =
    [ (t, c)
    | c <- everything
    , let ec = encodeContextCombo c
    , t <- allTags
    , encodeContextCombo (applyTagContextCombo t c) /= Context.applyTagContextCombo t ec ]

testSpaceCombos :: [ContextCombo]
testSpaceCombos =
    [ c
    | c <- everything
    , let ec = encodeContextCombo c
    , let esc = map (cc2w.encodeContextCombo) (spaceCombos c)
    , let sec = map cc2w (Context.spaceCombos ec)
    , sort esc /= sort sec ]

testSameSpaceContext :: [(Context, Context)]
testSameSpaceContext =
    [ (c1, c2)
    | c1 <- everything
    , let ec1 = encodeContext c1
    , c2 <- everything
    , let ec2 = encodeContext c2
    , sameSpaceContext c1 c2 /= Context.sameSpaceContext ec1 ec2 ]

testIsTT :: [ContextCombo]
testIsTT =
    [ c
    | c <- everything
    , isTT c /= Context.isTT (encodeContextCombo c) ]

testConsistency :: [(Context.ContextCombo, Context.ContextCombo,
                     Context.ContextCombo, Context.Context, Context.Cost)]
testConsistency =
    [ (c1, c2, c0, c00, cost)
    | c1 <- map encodeContextCombo everything
    , c2 <- map encodeContextCombo everything
    , Context.PF c0 cost <- Context.productionsFor c1 c2
    , c00 <- map encodeContext (comboToList (decodeContextCombo c0))
    , let (tags1, result1) = Context.pathTo c00 c1
    , let cost1 = Context.costTo c00 c1
    , let zero1 = Context.costTo result1 c1
    , let (tags2, result2) = Context.pathTo c00 c2
    , let cost2 = Context.costTo c00 c2
    , let zero2 = Context.costTo result2 c2
    , zero1 /= zeroCost || zero2 /= zeroCost ||
      cost1 /= encodeCost (costOfTags tags1) ||
      cost2 /= encodeCost (costOfTags tags2) ||
      cost1 +$ cost2 > cost ]
    where costOfTags tags
            = 7 * length tags
            + 2 * length (filter (`elem` [TTT, TEM, TPL]) tags)
