{-# OPTIONS -fglasgow-exts #-}

module Validate where

import PrelBase (Int(I#))
import PrelWord (Word(W#))
import Tag (Color, Tag(..))
import qualified Context
import Monad
import Maybe
import Bits
import Word
import Set

instance Show a => Show (Set a) where
    showsPrec _ x = showList' (setToList x) where
        showList' []     = showString "{}"
        showList' (x:xs) = showChar '{' . shows x . showl xs
                          where showl []     = showChar '}'
                                showl (x:xs) = showChar ',' . shows x .
                                               showl xs

data Switch = Off | On
    deriving (Eq, Ord, Show, Enum, Bounded)
data Enable = Enable { typewriter, bold, italic :: !Switch }
    deriving (Eq, Show)
data Underline = U0 | U1 | U2 | U3
    deriving (Eq, Ord, Show, Enum, Bounded)
data Volume = Base | EM | Strong
    deriving (Eq, Ord, Show, Enum, Bounded)
data VolumeSet = Simply !Volume | Weak
    deriving (Eq, Show)
data Minimize = Minimize !Enable !Underline !Volume
    deriving (Eq, Show)
data MinimizeSet = MinimizeSet !Enable !Underline !VolumeSet
    deriving (Eq, Show)

class (Eq a, Ord a) => Update a where
    decodeTag :: Tag -> Maybe a -- Nothing = irrelevant; Just = updating tag
    encodeTag :: a -> Tag
    bitList :: [(Rooted a, Int)]
type Size = Int
instance Update Int where       -- Update Size, really
    decodeTag (S size#) = Just (I# size#)
    decodeTag _         = Nothing
    encodeTag (I# size#) = S size#
    bitList = zip everything (10 : [0..9])
instance Update Color where
    decodeTag (C color) = Just color
    decodeTag _         = Nothing
    encodeTag = C
    bitList = zip everything (24 : [16..23])

data Rooted a = Root | Spec { unSpec :: !a }
    deriving (Eq, Ord, Show)

data Context = Context
    !Minimize !(Rooted Size) !(Rooted Color)
    deriving (Eq, Show)

type RootedSet a = Maybe (Rooted (Set a))
    -- Nothing       = we don't care at all what the field is, root or spec
    -- Just Root     = we require the field to be root
    -- Just (Spec _) = we require the field to be spec and in the given set

data ContextCombo = ContextCombo
    !MinimizeSet !(RootedSet Size) !(RootedSet Color)
    deriving (Eq, Show)

class Finite a where everything :: [a]
enumEverything :: (Enum a, Bounded a) => [a]
enumEverything = [minBound..maxBound]
instance Finite Switch where
  everything = enumEverything
instance Finite Enable where
  everything = [ Enable tt b i
               | tt <- everything, b <- everything, i <- everything ]
instance Finite Underline where
  everything = enumEverything
instance Finite Volume where
  everything = enumEverything
instance Finite VolumeSet where
  everything = Weak : map Simply everything
instance Finite Minimize where
  everything = [ Minimize e u v
               | e <- everything, u <- everything, v <- everything ]
instance Finite MinimizeSet where
  everything = [ MinimizeSet e u v
               | e <- everything, u <- everything, v <- everything ]
instance Finite Int where       -- really Finite Size
  everything = [0..9]
instance Finite Color where
  everything = enumEverything
instance Finite a => Finite (Rooted a) where
  everything = Root : map Spec everything
instance Finite Context where
  -- everything = [ Context m s c
  --              | m <- everything, s <- everything, c <- everything ]
  everything = [ Context m s c | m <- take 5 everything, s <- take 5 everything, c <- everything ]
            ++ [ Context m s c | m <- take 5 everything, s <- drop 5 everything, c <- take 5 everything ]
            ++ [ Context m s c | m <- drop 5 everything, s <- take 5 everything, c <- take 5 everything ]
instance Finite a => Finite (Maybe a) where
  everything = Nothing : map Just everything
instance (Eq a, Ord a, Finite a) => Finite (Set a) where
  -- everything = map (mkSet . concat) (sequence [ [[a],[]] | a <- everything ])
  -- The above is the really correct definition for "everything", but here
  -- we cheat and only pick "representative" samples to make things go faster.
  -- By "representative" we mean all 8 subsets of the first 3 candidate
  -- elements, and their complements, except the empty set.
  everything = filter (not . isEmptySet) (inc ++ map (all `minusSet`) inc)
      where inc = map (mkSet . concat)
                      (sequence [ [[a],[]] | a <- take 3 everything ])
            all = mkSet everything
instance Finite ContextCombo where
  -- everything = [ ContextCombo m s c
  --              | m <- everything, s <- everything, c <- everything ]
  everything = [ ContextCombo m s c | m <- take 5 everything, s <- take 5 everything, c <- everything ]
            ++ [ ContextCombo m s c | m <- take 5 everything, s <- drop 5 everything, c <- take 5 everything ]
            ++ [ ContextCombo m s c | m <- drop 5 everything, s <- take 5 everything, c <- take 5 everything ]

type Cost = Int

w2c :: Word -> Context.Context
w2c (W# w#) = Context.Context# w#
w2cc :: Word -> Context.ContextCombo
w2cc (W# w#) = Context.ContextCombo# w#
c2w :: Context.Context -> Word
c2w (Context.Context# w#) = W# w#
cc2w :: Context.ContextCombo -> Word
cc2w (Context.ContextCombo# w#) = W# w#

encodeContext :: Context -> Context.Context
encodeContext (Context m s c) = w2c
    (encodeMinimize m .|. encodeUpd s .|. encodeUpd c)
encodeContextCombo :: ContextCombo -> Context.ContextCombo
encodeContextCombo (ContextCombo m s c) = w2cc
    (encodeMinimizeSet m .|. encodeUpdSet s .|. encodeUpdSet c)
decodeContext :: Context.Context -> Context
decodeContext w = Context
    (decodeMinimize w') (decodeUpd w') (decodeUpd w')
    where w' = c2w w
decodeContextCombo :: Context.ContextCombo -> ContextCombo
decodeContextCombo w = ContextCombo
    (decodeMinimizeSet w') (decodeUpdSet w') (decodeUpdSet w')
    where w' = cc2w w

encodeMinimize :: Minimize -> Word
encodeMinimize (Minimize e u v)
    = encodeEnable e .|. encodeUnderline u .|. encodeVolume v
decodeMinimize :: Word -> Minimize
decodeMinimize w
    = Minimize (decodeEnable w) (decodeUnderline w) (decodeVolume w)
encodeMinimizeSet :: MinimizeSet -> Word
encodeMinimizeSet (MinimizeSet e u v)
    = encodeEnable e .|. encodeUnderline u .|. encodeVolumeSet v
decodeMinimizeSet :: Word -> MinimizeSet
decodeMinimizeSet w
    = MinimizeSet (decodeEnable w) (decodeUnderline w) (decodeVolumeSet w)

encodeEnable :: Enable -> Word
encodeEnable (Enable tt b i)
    =   (if tt == On then bit 31 else 0)
    .|. (if b  == On then bit 30 else 0)
    .|. (if i  == On then bit 29 else 0)
decodeEnable :: Word -> Enable
decodeEnable w = Enable
    (if testBit w 31 then On else Off)
    (if testBit w 30 then On else Off)
    (if testBit w 29 then On else Off)

encodeUnderline :: Underline -> Word
encodeUnderline u = shiftL n 25
    where n = case u of { U0 -> 0; U1 -> 1; U2 -> 2; U3 -> 3 }
decodeUnderline :: Word -> Underline
decodeUnderline w
    = case [ u | u <- everything
               , w .&. encodeUnderline U3 == encodeUnderline u ]
      of [u] -> u

encodeVolume :: Volume -> Word
encodeVolume Base   = shiftL 0 27
encodeVolume EM     = shiftL 3 27
encodeVolume Strong = shiftL 2 27
decodeVolume :: Word -> Volume
decodeVolume w
    = case [ v | v <- everything
               , w .&. encodeVolume EM == encodeVolume v ]
      of [v] -> v
encodeVolumeSet :: VolumeSet -> Word
encodeVolumeSet (Simply v) = encodeVolume v
encodeVolumeSet (Weak    ) = shiftL 1 27
decodeVolumeSet :: Word -> VolumeSet
decodeVolumeSet w
    = if w .&. encodeVolume EM == encodeVolumeSet Weak
      then Weak else Simply (decodeVolume w)

{-# SPECIALIZE encodeUpd :: Rooted Color -> Word #-}
{-# SPECIALIZE encodeUpd :: Rooted Size  -> Word #-}
{-# SPECIALIZE decodeUpd :: Word -> Rooted Color #-}
{-# SPECIALIZE decodeUpd :: Word -> Rooted Size  #-}
{-# SPECIALIZE encodeUpdSet :: RootedSet Color -> Word #-}
{-# SPECIALIZE encodeUpdSet :: RootedSet Size  -> Word #-}
{-# SPECIALIZE decodeUpdSet :: Word -> RootedSet Color #-}
{-# SPECIALIZE decodeUpdSet :: Word -> RootedSet Size  #-}
encodeUpd :: Update a => Rooted a -> Word
encodeUpd = toBits
decodeUpd :: Update a => Word -> Rooted a
decodeUpd w = case fromBits w of [a] -> a
encodeUpdSet :: Update a => RootedSet a -> Word
encodeUpdSet rs@Nothing = foldl1 (.|.) [ bit i | (a,i) <- bitList' ]
    where bitList' = bitList `asTypeOf`
            zip (map Spec (setToList (unSpec (fromJust rs)))) []
encodeUpdSet (Just rs@Root) = encodeUpd root
    where root = Root `asTypeOf` Spec (head (setToList (unSpec rs)))
encodeUpdSet (Just (Spec set)) = foldl (.|.) 0 words
    where words = map (encodeUpd.Spec) (setToList set)
decodeUpdSet :: Update a => Word -> RootedSet a
decodeUpdSet w = rs2rs (mkSet (fromBits w))

{-# SPECIALIZE toBits :: Rooted Color -> Word #-}
{-# SPECIALIZE toBits :: Rooted Size  -> Word #-}
{-# SPECIALIZE fromBits :: Word -> [Rooted Color] #-}
{-# SPECIALIZE fromBits :: Word -> [Rooted Size ] #-}
{-# SPECIALIZE rs2rs :: Set (Rooted Color) -> RootedSet Color #-}
{-# SPECIALIZE rs2rs :: Set (Rooted Size ) -> RootedSet Size  #-}
toBits :: (Update a, Bits b) => Rooted a -> b
toBits a = bit (fromJust (lookup a bitList))
fromBits :: (Update a, Bits b) => b -> [Rooted a]
fromBits b = [ a | (a,i) <- bitList, testBit b i ]
rs2rs :: Update a => Set (Rooted a) -> RootedSet a
rs2rs set
    | hasRoot && set == unitSet Root
    = Just Root
    | hasRoot && set == mkSet (map fst bitList)
    = Nothing
    | not hasRoot
    = Just (Spec (mapSet unSpec set))
    where hasRoot = Root `elementOf` set

productionsFor :: ContextCombo -> ContextCombo -> [(ContextCombo, Cost)]
productionsFor (ContextCombo m1 s1 c1) (ContextCombo m2 s2 c2)
    = [ (ContextCombo m0 s0 c0, mC + sC + cC)
      | (m0, mC) <- productionsForM m1 m2
      , (s0, sC) <- productionsForUpd s1 s2
      , (c0, cC) <- productionsForUpd c1 c2 ]

productionsForM :: MinimizeSet -> MinimizeSet -> [(MinimizeSet, Cost)]
productionsForM m1 m2
    = noPL : ( (if min == m2 then id else l_PL)
             $ (if min == m1 then id else r_PL) [] )
    where noPL @ (min, cost) = minAndCost m1 m2
          l_PL = ((m2, priorCostM m1 + 9) :)
          r_PL = ((m1, priorCostM m2 + 9) :)

minAndCost :: MinimizeSet -> MinimizeSet -> (MinimizeSet, Cost)
minAndCost (MinimizeSet e1 u1 v1) (MinimizeSet e2 u2 v2)
    = (MinimizeSet e0 u0 v0, eC + uC + vC)
    where (e0, eC) = minAndCostE e1 e2
          (u0, uC) = minAndCostU u1 u2
          (v0, vC) = minAndCostV v1 v2
minAndCostE (Enable tt1 b1 i1) (Enable tt2 b2 i2)
    = ( Enable (if eqTT then tt1 else Off)
               (if eqB  then b1  else Off)
               (if eqI  then i1  else Off)
      , (if eqTT then 0 else 9) +
        (if eqB  then 0 else 7) +
        (if eqI  then 0 else 7) )
    where eqTT = tt1 == tt2
          eqB  = b1  == b2
          eqI  = i1  == i2
minAndCostU u1 u2 | u1 < u2   = (u1, 7 * (fromEnum u2 - fromEnum u1))
minAndCostU u1 u2 | otherwise = (u2, 7 * (fromEnum u1 - fromEnum u2))
minAndCostV (Simply Strong) (Simply Strong) = (Simply Strong, 0)
minAndCostV (Simply Strong) other           = (other        , 7)
minAndCostV other           (Simply Strong) = (other        , 7)
minAndCostV Weak            other           = (other        , 0)
minAndCostV other           Weak            = (other        , 0)
minAndCostV (Simply Base)   (Simply Base)   = (Simply Base  , 0)
minAndCostV (Simply EM)     (Simply EM)     = (Simply EM    , 0)
minAndCostV _               _               = (Weak         , 9)

{-# SPECIALIZE productionsForUpd :: RootedSet Color -> RootedSet Color -> [(RootedSet Color, Cost)] #-}
{-# SPECIALIZE productionsForUpd :: RootedSet Size  -> RootedSet Size  -> [(RootedSet Size , Cost)] #-}
productionsForUpd :: Update a =>
    RootedSet a -> RootedSet a -> [(RootedSet a, Cost)]
productionsForUpd Nothing rs = [(rs, 0)]
productionsForUpd rs Nothing = [(rs, 0)]
productionsForUpd (Just Root) (Just Root) = [(Just Root, 0)]
productionsForUpd (Just Root) (Just (Spec _)) = [(Just Root, 7)]
productionsForUpd (Just (Spec _)) (Just Root) = [(Just Root, 7)]
productionsForUpd (Just (Spec set1)) (Just (Spec set2))
    = [ (Just (Spec set0), 0)
      | let set0 = intersect set1 set2
      , not (isEmptySet set0) ]
   ++ [ (Just (Spec set0), 7)
      | let set0 = union (minusSet set1 set2) (minusSet set2 set1)
      , not (isEmptySet set0) ]

pathTo :: Context -> ContextCombo -> ([Tag], Context)
pathTo (Context m0 s0 c0) (ContextCombo m1 s1 c1)
    = (mTags ++ cTags ++ sTags, Context mResult sResult cResult)
    -- We harmlessly swap the order of cTags and sTags above
    -- so as to match the implementation in Context.
    where (mTags, mResult) = pathToM   m0 m1
          (sTags, sResult) = pathToUpd s0 s1
          (cTags, cResult) = pathToUpd c0 c1

pathToM m0@(Minimize    (Enable tt0 b0 i0) u0 v0)
        m1@(MinimizeSet (Enable tt1 b1 i1) u1 v1)
    = if tt0 > tt1
      || b0  > b1
      || i0  > i1
      || u0  > u1
      || v0 == Strong && v1 /= Simply Strong
      then case pathToM' rootContextM m1 of
           (mTags, mResult) -> (TPL:mTags, mResult)
      else pathToM' m0 m1
pathToM' (Minimize    e0 u0 v0)
         (MinimizeSet e1 u1 v1)
    = (eTags ++ uTags ++ vTags, Minimize eResult uResult vResult)
    where (eTags, eResult) = pathToM'E e0 e1
          (uTags, uResult) = pathToM'U u0 u1
          (vTags, vResult) = pathToM'V v0 v1
pathToM'E (Enable tt0 b0 i0) e1@(Enable tt1 b1 i1)
    = ([TTT | tt0 < tt1] ++ [TB | b0 < b1] ++ [TI | i0 < i1], e1)
pathToM'U u0 u1 = (replicate (fromEnum u1 - fromEnum u0) TU, u1)
pathToM'V Strong (Simply Strong) = ([], Strong)
pathToM'V _      (Simply Strong) = ([TS], Strong)
pathToM'V Strong _               = error "Cannot undo <S> without <PL>"
pathToM'V v0     Weak            = ([], v0)
pathToM'V v0     (Simply v1)     = ([TEM | v0 /= v1], v1)

{-# SPECIALIZE pathToUpd :: Rooted Color -> RootedSet Color -> ([Tag], Rooted Color) #-}
{-# SPECIALIZE pathToUpd :: Rooted Size  -> RootedSet Size  -> ([Tag], Rooted Size ) #-}
pathToUpd :: Update a => Rooted a -> RootedSet a -> ([Tag], Rooted a)
pathToUpd old          Nothing           = ([], old)
pathToUpd Root         (Just Root)       = ([], Root)
pathToUpd old@(Spec a) (Just (Spec set)) | a `elementOf` set = ([], old)
pathToUpd _            (Just (Spec set)) = ([encodeTag target], Spec target)
                                            where target = head (setToList set)

unitCombo :: Context -> ContextCombo
unitCombo (Context m s c) = ContextCombo
    (unitComboM m) (unitComboUpd s) (unitComboUpd c)

unitComboM :: Minimize -> MinimizeSet
unitComboM (Minimize e u v) = MinimizeSet e u (Simply v)

unitComboUpd :: Update a => Rooted a -> RootedSet a
unitComboUpd (Root  ) = Just (Root            )
unitComboUpd (Spec a) = Just (Spec (unitSet a))

rootContext :: Context
rootContext = Context rootContextM Root Root

rootContextM :: Minimize
rootContextM = Minimize (Enable Off Off Off) U0 Base

priorCost :: ContextCombo -> Cost
priorCost (ContextCombo m s c) = priorCostM m + priorCostUpd s + priorCostUpd c

priorCostM (MinimizeSet e u v) = priorCostE e + priorCostU u + priorCostV v
priorCostE (Enable tt b i)
    = (if tt == On then 9 else 0)
    + (if b  == On then 7 else 0)
    + (if i  == On then 7 else 0)
priorCostU U0 = 0
priorCostU U1 = 7
priorCostU U2 = 14
priorCostU U3 = 21
priorCostV (Simply Base  ) = 0
priorCostV (Simply EM    ) = 9
priorCostV (Simply Strong) = 7
priorCostV (Weak         ) = 0

priorCostUpd :: Update a => RootedSet a -> Cost
priorCostUpd (Nothing      ) = 0
priorCostUpd (Just (Root  )) = 0
priorCostUpd (Just (Spec _)) = 7

applyTagContext :: Tag -> Context -> Context
applyTagContext tag (Context m s c) = Context
    (applyTagMinimize tag m) (applyTagUpd tag s) (applyTagUpd tag c)
applyTagContextCombo :: Tag -> ContextCombo -> ContextCombo
applyTagContextCombo tag (ContextCombo m s c) = ContextCombo
    (applyTagMinimizeSet tag m) (applyTagUpdSet tag s) (applyTagUpdSet tag c)
applyTagMinimize :: Tag -> Minimize -> Minimize
applyTagMinimize tag (Minimize e u v) = Minimize
    (applyTagEnable tag e) (applyTagUnderline tag u) (applyTagVolume tag v)
applyTagMinimizeSet :: Tag -> MinimizeSet -> MinimizeSet
applyTagMinimizeSet tag (MinimizeSet e u v) = MinimizeSet
    (applyTagEnable tag e) (applyTagUnderline tag u) (applyTagVolumeSet tag v)
applyTagEnable :: Tag -> Enable -> Enable
applyTagEnable TTT (Enable tt b i) = Enable On b i
applyTagEnable TB  (Enable tt b i) = Enable tt On i
applyTagEnable TI  (Enable tt b i) = Enable tt b On
applyTagEnable TPL _               = Enable Off Off Off
applyTagEnable _   e               = e
applyTagUnderline :: Tag -> Underline -> Underline
applyTagUnderline TU  U0 = U1
applyTagUnderline TU  U1 = U2
applyTagUnderline TU  U2 = U3
applyTagUnderline TPL _  = U0
applyTagUnderline _   u  = u
applyTagVolume :: Tag -> Volume -> Volume
applyTagVolume TEM Base = EM
applyTagVolume TEM EM   = Base
applyTagVolume TS  _    = Strong
applyTagVolume TPL _    = Base
applyTagVolume _   v    = v
applyTagVolumeSet :: Tag -> VolumeSet -> VolumeSet
applyTagVolumeSet tag (Simply v) = Simply (applyTagVolume tag v)
applyTagVolumeSet tag Weak
    = case (applyTagVolume tag Base, applyTagVolume tag EM) of
      (Base  , Base  ) -> Simply Base
      (EM    , EM    ) -> Simply EM
      (Strong, Strong) -> Simply Strong
      (Base  , EM    ) -> Weak
      (EM    , Base  ) -> Weak
{-# SPECIALIZE applyTagUpd :: Tag -> Rooted Color -> Rooted Color #-}
{-# SPECIALIZE applyTagUpd :: Tag -> Rooted Size  -> Rooted Size  #-}
{-# SPECIALIZE applyTagUpdSet :: Tag -> RootedSet Color -> RootedSet Color #-}
{-# SPECIALIZE applyTagUpdSet :: Tag -> RootedSet Size  -> RootedSet Size  #-}
applyTagUpd :: Update a => Tag -> Rooted a -> Rooted a
applyTagUpd tag old
    = case decodeTag tag of
      Just new -> Spec new
      Nothing  -> old
applyTagUpdSet :: Update a => Tag -> RootedSet a -> RootedSet a
applyTagUpdSet tag old
    = case decodeTag tag of
      Just new -> Just (Spec (unitSet new))
      Nothing  -> old

spaceCombos :: ContextCombo -> [ContextCombo]
spaceCombos (ContextCombo (MinimizeSet (Enable tt _ _) u _) s c)
    = [ ContextCombo (MinimizeSet (Enable tt b' i') u v') s c'
      | b' <- [Off, On]
      , i' <- [Off, On]
      , v' <- [Weak, Simply Strong] ]
    where c' = if u == U0 then Nothing else c

sameSpaceContext :: Context -> Context -> Bool
sameSpaceContext (Context (Minimize (Enable tt1 _ _) u1 _) s1 c1)
                 (Context (Minimize (Enable tt2 _ _) u2 _) s2 c2)
    = (u1 == u2) && (u1 == U0 || c1 == c2) && (tt1 == tt2) && (s1 == s2)

isTT :: ContextCombo -> Bool
isTT (ContextCombo (MinimizeSet (Enable tt _ _) _ _) _ _) = tt == On
